####
# bomail.cli.send
#
# Send drafts.
####

import os
import sys
import mimetypes
import subprocess, email
import smtplib
import traceback

from email import encoders
from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart
from email.mime.audio import MIMEAudio
from email.mime.base import MIMEBase
from email.mime.image import MIMEImage

from bomail.config.config import pathcfg,sendcfg
import bomail.cli.mailfile as mailfile
import bomail.util.addr as addr
import bomail.util.util as util

usage_str = """
Reads filename(s) from stdin one per line and sends them.
send.py -h to print this help.
"""


# adapted from python library example: https://docs.python.org/3/library/email-examples.html
def add_attachments(msg, attach_list):
  for filename in attach_list:
    if not os.path.exists(filename):
      raise Exception("Did not send. Could not attach (does not exist): " + filename)
    if not os.path.isfile(filename):
      raise Exception("Did not send. Could not attach (is not a file): " + filename)
    ctype, encoding = mimetypes.guess_type(filename)
    if ctype is None or encoding is not None:
      ctype = 'application/octet-stream'
    maintype, subtype = ctype.split('/', 1)
    if maintype == 'text':
      with open(filename) as fp:
        sub_msg = MIMEText(fp.read(), _subtype=subtype)
    elif maintype == 'image':
      with open(filename, 'rb') as fp:
        sub_msg = MIMEImage(fp.read(), _subtype=subtype)
    elif maintype == 'audio':
      with open(filename, 'rb') as fp:
        sub_msg = MIMEAudio(fp.read(), _subtype=subtype)
    else:
      with open(filename, 'rb') as fp:
        sub_msg = MIMEBase(maintype, subtype)
        sub_msg.set_payload(fp.read())
      encoders.encode_base64(sub_msg)
    sub_msg.add_header('Content-Disposition', 'attachment', filename=os.path.split(filename)[1])
    msg.attach(sub_msg)


def file_to_msg(fname, mgr):
  data = mgr.get_all(fname)
  attach_list = mgr.get_attachlist(fname)
  if len(attach_list) == 0:  # no attachments
    msg = MIMEText(data[mailfile.BODY_L], 'plain')
  else:
    msg = MIMEMultipart()
    msg.attach(MIMEText(data[mailfile.BODY_L], 'plain'))
    add_attachments(msg, attach_list)
  msg['Subject'] = data[mailfile.SUBJ_L]
  msg['From'] = data[mailfile.FROM_L]
  msg['To'] = data[mailfile.TO_L]
  msg['Date'] = email.utils.formatdate()
  msg['Message-ID'] = data[mailfile.MSG_ID_L]
  if len(data[mailfile.CC_L]) > 0:
    msg['CC'] = data[mailfile.CC_L]
  refs = data[mailfile.REFS_L]
  if len(data[mailfile.REPLY_L]) > 0:
    msg['Reply-To'] = data[mailfile.REPLY_L]
  if len(refs) > 0:
    msg['In-Reply-To'] = refs.split(", ")[-1]
    msg['References'] = refs

  recip_lists = data[mailfile.TO_L].split(", ")
  if len(data[mailfile.CC_L]) > 0:
    recip_lists += data[mailfile.CC_L].split(", ")
  if len(data[mailfile.BCC_L]) > 0:
    recip_lists += data[mailfile.BCC_L].split(", ")
  recip_addrs = [addr.str_to_pair(a)[1] for a in recip_lists]
  return recip_addrs, msg


# given list of relative filenames
# return results, msg
# where results is a list of success/not [True, False, ...]
def connect_and_send(email_addr, flist, mgr):
  results = [False for f in flist]
  try:
    user_ind = sendcfg.email_addrs.index(email_addr)
  except:
    return results, "Error sending from " + email_addr + " (not one of your addresses)"

  try:
    if sendcfg.smtp_servernames[user_ind] == "localhost":
      serv = smtplib.SMTP("localhost")
    else:
      serv = smtplib.SMTP(sendcfg.smtp_servernames[user_ind], sendcfg.smtp_ports[user_ind])
      serv.ehlo()
      serv.starttls()
      serv.ehlo()
      serv.login(sendcfg.smtp_usernames[user_ind], sendcfg.smtp_passwords[user_ind])
  except Exception as e:
    util.err_log("Sending email, error connecting to server.\n" + traceback.format_exc() + "\n")
    return results, "Error connecting to server: " + str(e)

  for i, fname in enumerate(flist):
    recip_addrs, msg = file_to_msg(fname, mgr)
    try:
      serv.sendmail(email_addr, recip_addrs, msg.as_string())
      if sendcfg.smtp_servernames[user_ind] == "localhost": # save the sent mail
        os.makedirs(pathcfg.new_rawmail_dir, exist_ok=True)
        try:
          rawname = util.mime_to_mailfile.get_msgid(m, fname)
          with open(os.path.join(pathcfg.new_rawmail_dir,rawname), "w") as rawfile:
            rawfile.write(msg.as_string())
        except:
          util.err_log("Error saving copy of sent mail:\n" + fname + "\n" + msg.as_string() + "\n")
      results[i] = True
    except Exception as e:
      serv.quit()
      util.err_log("Sending email, error while sending.\n" + traceback.format_exc() + "\n")
      return results, "Error sending message #" + str(i) + " [" + fname + "]: " + str(e)

  serv.quit()
  return results, ""

 

# given list of filenames
# return vector of successes [True, False, ...], err_msg
def main(flist, mgr):
  # sort them into different "from" addresses and connect to each server once
  fromaddr_to_fnames = {}
  for fname in flist:
    fromline = mgr.get(fname, mailfile.FROM_L)
    frompr = addr.str_to_pair(fromline)
    fromaddr = frompr[1]
    if fromaddr not in fromaddr_to_fnames:
      fromaddr_to_fnames[fromaddr] = [fname]
    else:
      fromaddr_to_fnames[fromaddr].append(fname)

  success = set()
  msg = ""
  for fromaddr,fnames in fromaddr_to_fnames.items():
    res, msg = connect_and_send(fromaddr, fnames, mgr)
    for i,good in enumerate(res):
      if good:
        success.add(fnames[i])
    if len(msg) > 0:
      break
  return [f in success for f in flist], msg



def main_cli():
  if len(sys.argv) >= 2:
    print(usage_str)
    exit(0)

  flist = [f.strip() for f in sys.stdin.readlines()]
  res, err = main(flist, mailfile.MailMgr())
  print("Sent " + str(res) + "/" + str(len(flist)) + " messages.")
  if res < len(flist):
    print(err)


if __name__ == "__main__":
  main_cli()





####
# bomail.cli.compose
#
# Compose new emails as blank, reply, forward.
####

import sys
import string
import traceback
import os, random
import dateutil, dateutil.parser
import email, email.utils, uuid

from bomail.config.config import sendcfg,optcfg
import bomail.cli.mailfile as mailfile
import bomail.cli.chstate as chstate
import bomail.util.addr as addr
import bomail.util.util as util
import bomail.util.datestuff as datestuff

usage_str = """Use with option -h to print this help.
There are two forms of usage.

1. To compose a new email from scratch, use no options or any combination of:
  -to "example here"       add "example string" to recips (may be comma-sep list)
  -subj "example here"     set subject line to "example here"
  -cc "example here"       add "example here" to cc (may be comma-sep list)
  -bcc "example here"      ditto for bcc
  -reply "example here"    ditto for reply-to

2. To compose based on an existing email, use one of:
  -ro filename             reply-one to msgid
  -ra filename             reply-all to msgid
  -rr filename             reply-recipients to msgid
  -f  filename "example"   forward msgid to "example"

(Both 1 and 2.) Optional, add:
  -body "body of email here"        
"""


# for a draft, just set sent and date
def create(data, mgr):
  data[mailfile.DATE_L] = datestuff.get_local_nowstr()
  fname = mailfile.get_filename(data, True)  # is draft
  chstate.create([fname], [data], mgr)
  return fname


def get_blank_draft_data():
  new_data = [""]*len(mailfile.fields)
  new_data[mailfile.FROM_L] = addr.pair_to_str(sendfg.names[0], sendcfg.email_addrs[0])
  new_data[mailfile.DATE_L] = datestuff.get_local_nowstr()
  new_data[mailfile.SENT_L] = "True"
  new_data[mailfile.STATE_L] = "open"
  # create a message id, at least for now
  new_data[mailfile.MSG_ID_L] = email.utils.make_msgid(uuid.uuid4().hex)
  return new_data


# for a reply or forward, build new data from the old one
# if quote, include previous email message as quoted text
def get_old_and_new_data(fname, quote):
  old_data = mailfile.read_file(fname)
  new_data = get_blank_draft_data()
  # If user has multiple addresses: attempt to reply from the address it is addressed to
  for name,email_addr in reversed(list(zip(sendcfg.names,sendcfg.email_addrs))):
    if any([email_addr in old_data[j] for j in [mailfile.TO_L, mailfile.CC_L,mailfile.BCC_L]]):
      new_data[mailfile.FROM_L] = addr.pair_to_str(name, email_addr)
  new_data[mailfile.REFS_L] = old_data[mailfile.MSG_ID_L] + ", " + old_data[mailfile.REFS_L]
  new_data[mailfile.TAGS_L] = old_data[mailfile.TAGS_L] if optcfg.autotag_draft_replies else ""
  prev_date_obj = dateutil.parser.parse(old_data[mailfile.DATE_L])
  prev_date_str = prev_date_obj.strftime(optcfg.quote_date_fmt)
  prev_time_str = prev_date_obj.strftime(optcfg.quote_time_fmt)
  if quote:
    body_lines = old_data[mailfile.BODY_L].split("\n")
    body_str = optcfg.quote_line_prefix + ("\n" + optcfg.quote_line_prefix).join(body_lines)
    new_data[mailfile.BODY_L] = optcfg.quote_format.replace("%date", prev_date_str).replace("%time", prev_time_str).replace("%from", old_data[mailfile.FROM_L]).replace("%body", body_str)
  return old_data, new_data


# get the recipient(s) to reply to
def get_reply_to(data):
  if len(data[mailfile.REPLY_L]) > 0:
    return data[mailfile.REPLY_L]
  else:
    return data[mailfile.FROM_L]

  
# prepend Re: or Fwd: if not already prepended
def get_subject_with(old_sub, prefix):
  if old_sub[:len(prefix)].lower() == prefix.lower():
    return old_sub
  else:
    return prefix + old_sub


def reply_one(fname, mgr, quote):
  old_data, new_data = get_old_and_new_data(fname, quote)
  new_data[mailfile.SUBJ_L] = get_subject_with(old_data[mailfile.SUBJ_L], "Re: ")
  new_data[mailfile.TO_L] = get_reply_to(old_data)
  return create(new_data, mgr)


def reply_all(fname, mgr, quote, include_sender=True):
  old_data, new_data = get_old_and_new_data(fname, quote)
  new_data[mailfile.SUBJ_L] = get_subject_with(old_data[mailfile.SUBJ_L], "Re: ")
  to_list = [s for s in old_data[mailfile.TO_L].split(", ") if not addr.is_str_me(s)]
  to_list += [s for s in old_data[mailfile.CC_L].split(", ") if not addr.is_str_me(s)]
  if include_sender:
    to_list = get_reply_to(old_data).split(", ") + to_list
  new_data[mailfile.TO_L] = ", ".join(to_list)
  return create(new_data, mgr)


def forward(fname, to_str, mgr, quote=True):
  old_data, new_data = get_old_and_new_data(fname, quote)
  new_data[mailfile.SUBJ_L] = get_subject_with(old_data[mailfile.SUBJ_L], "Fwd: ")
  new_data[mailfile.TO_L] = to_str
  return create(new_data, mgr)


def new_compose(subj_str, to_str, cc_str, bcc_str, reply_str, bodystr, mgr):
  d = get_blank_draft_data()
  d[mailfile.TO_L] = to_str
  d[mailfile.CC_L] = cc_str
  d[mailfile.BCC_L] = bcc_str
  d[mailfile.REPLY_L] = reply_str
  d[mailfile.SUBJ_L] = subj_str
  d[mailfile.BODY_L] = bodystr
  return create(d, mgr)


def main_cli():
  if "-h" in sys.argv:
    print(usage_str)
    exit(0)
  try:
    args = sys.argv
    bodystr = ""
    mgr = mailfile.MailMgr()
    if len(args) <= 1:
      print(new_compose("", "", "", "", "", mgr))
      return

    if "-body" in args:
      ind = args.index("-body")
      if ind < len(args) - 1:
        bodystr = args[ind+1]
        args = args[:ind] + args[ind+2:]

    if args[1] == "-ro":
      print(reply_one(args[2], bodystr, mgr))
    elif args[1] == "-ra":
      print(reply_all(args[2], bodystr, mgr))
    elif args[1] == "-rr":
      print(reply_all(args[2], bodystr, mgr, include_sender=False))
    elif args[1] == "-f":
      print(forward(args[2], args[3], bodystr, mgr, quote=True))
    else:
      to_str_list = []
      cc_str_list = []
      bcc_str_list = []
      reply_str_list = []
      subj_str = ""
      i = 1
      while i < len(args):
        if args[i] == "-to":
          to_str_list.append(args[i+1])
          i += 1
        elif args[i] == "-subj":
          subj_str = args[i+1]
          i += 1
        elif args[i] == "-cc":
          cc_str_list.append(args[i+1])
          i += 1
        elif args[i] == "-bcc":
          bcc_str_list.append(args[i+1])
          i += 1
        elif args[i] == "-reply":
          reply_str_list.append(args[i+1])
          i += 1
        i += 1
      print(new_compose(subj_str, ", ".join(to_str_list), ", ".join(cc_str_list), ", ".join(bcc_str_list), ", ".join(reply_str_list), bodystr, mgr))
  except Exception:
    traceback.print_exc()
    print("")
    print(usage_str)


if __name__ == "__main__":
  main_cli()




####
# bomail.cli.mailfile
#
# Specifies the bomail file format, called 'mailfile',
# and mail manager object, and some commands to view/edit
# metadata of emails.
####

import sys
import os, bisect
import email.utils, uuid
import traceback

from bomail.config.config import pathcfg
import bomail.util.tags as tags
import bomail.util.attach as attach
import bomail.util.msgids as msgids
import bomail.util.datestuff as datestuff
import bomail.util.util as util

####
# This file specifies
# (1) the mailfile format and interface
# (2) MailMgr, the mail manager object
#
# Emails are stored in "mailfiles".
# A mailfile is of the form:
#   field_name: field_value
#   ...
#   field_name: field_value
#   =================
#   body_of_email
#
# Formats of specific entries:
# "addr" can be either of these:
#    name@domain.com
#    Longer Name Here <name@domain.com>
#
# "msg_id" is an email message id enclosed in < >
#
# "date" is an ISO8061 date string in UTC time, i.e. yyyy-mm-ddTHH:MM:SS or some prefix
#
# attachment file paths are quote-enclosed
# (with inner quotes backslash-escaped, and all backslashes escaped)
#
# no newlines except in the body
#
# all lists are comma-and-space separated, implying that their
# elements shouldn't have either commas or spaces in them
# except attachment filenames which are quote enclosed
####

fields = ["from","to","cc","bcc","reply-to","subject","date","sent","msg-id","references","referenced_by","attachments","state","tags","body"]
FROM_L   = 0    # addr
TO_L     = 1    # addr list
CC_L     = 2    # addr list
BCC_L    = 3    # addr list
REPLY_L  = 4    # addr list (reply-to header)
SUBJ_L   = 5    # string with no newline
DATE_L   = 6    # datestr (ISO8061 format)
SENT_L   = 7    # "True" if sent from me, "False" otherwise
MSG_ID_L = 8    # msg_id
REFS_L   = 9    # list of msg_ids
REF_BY_L = 10   # list of msg_ids
ATTACH_L = 11   # list of attachment files
STATE_L  = 12   # "open", "closed", or "scheduled datestr"
TAGS_L   = 13   # list of tags
BODY_L   = 14   # arbitrary string

sep = "================\n"  # between headers and body

usage_str = """
Usage: reads a list of filenames from stdin and does the action for all of them.
Options: -g (get the indicated fields), -add-tags, -rm-tags

  mailfile.py -h                  # print this help
  mailfile.py -g all              # get and print all fields
  mailfile.py -g from bcc date    # get and print the fields 'from', 'bcc', and 'date'
  mailfile.py -add-tags tag1 tag2 # add the tags 'tag1' and 'tag2'
  mailfile.py -rm-tags tag1 tag2  # remove the tags 'tag1' and 'tag2'
"""


error_data = [
  "error file not found <>", "", "", "", "",  # from, to, cc, bcc, reply-to
  "", "9999-12-28", "False", email.utils.make_msgid(uuid.uuid4().hex), # subj, date, sent, msgid
  "", "", "", "open", "", # refs, ref-by, attach, state, tags
  "error file not found"]  # body



#-------------------------------
# fileid is of the form yyyy/mm-dd/THH:MM-msgid.ext

DRAFT_EXT = ".draft"
MAIL_EXT = ".mail"
MIME_EXT = ".mime"

def msgid_as_path(msg_id):
  return msg_id.replace("/","-")

def get_datestart_file_part(data):
  msgid = data[MSG_ID_L]
  # convert date to utc
  datestr = data[DATE_L]
  obj = datestuff.parse_to_utc_obj(datestr)
  s = obj.strftime("%Y/%m-%d/T%H:%M:%S")    # maybe not cross-platform
  return s + "-" + msgid_as_path(msgid)

def get_datestart_filename(data, is_draft):
  s = get_datestart_file_part(data)
  ext = DRAFT_EXT if is_draft else MAIL_EXT
  return s + ext

# return full path to mailfile
def get_filename(data, is_draft):
  s = get_datestart_filename(data, is_draft)
  d = pathcfg.drafts_dir if is_draft else pathcfg.email_dir
  return os.path.join(d, s)

# where its attachments are stored
def get_attach_dir(data, is_draft):
  s = get_datestart_filename(data, is_draft)
  return os.path.join(pathcfg.attach_dir, s)

def get_trash_attach_dir(data, is_draft):
  s = get_datestart_filename(data, is_draft)
  return os.path.join(pathcfg.trash_dir, "attach", s)

# return full file path
def get_trashfilename(data, is_draft):
  s = get_datestart_filename(data, is_draft)
  if is_draft:
    return os.path.join(pathcfg.trash_dir, "draft", s)
  return os.path.join(pathcfg.trash_dir, s)

# return fullfile path
def get_rawfilename(data):
  s = get_datestart_file_part(data)
  return os.path.join(pathcfg.old_rawmail_dir, s + MIME_EXT)


#--------------------------------------
# reading/writing to files

# filename is full path to file
# return data array
def read_file(filename):
  data = []
  try:
    with open(filename) as f:
      for i in range(len(fields)-1):
        line = f.readline()
        data.append(line[len(fields[i]) + 2 : -1])  # erase "field: " and trailing "\n"
      f.readline()  # separator ==========
      data.append(f.read()) # body
  except:
    util.err_log("mailfile: could not find " + str(filename) + "\n")
    return list(error_data)
  return data

# filename is a full path
def write_to_file(filename, data):
  os.makedirs(filename[:filename.rfind("/")], exist_ok=True)
  with open(filename, "w") as f:  # if not, crash because that's bad!!
    f.write(data_to_str(data))


#-------------------------------
# formatting of things

# turn refstr into list of msgids
# if my_id is not None, make sure to remove it from refs
# if it is None, ignore it
def do_get_referencelist(my_id, refstr):
  reflist = []
  if len(refstr) == 0:
    return reflist
  for r in refstr.split(", "):
    rs = r.strip()
    if len(rs) > 0 and (my_id is None or rs != my_id):
      reflist.append(rs)
  return reflist


def do_get_parent_id(my_id, reflist):
  if len(reflist) == 0 or reflist[-1] == my_id:
   return None
  return reflist[-1]


def data_to_str(data):
  a = []
  for i,field in enumerate(fields[:-1]):
    a.append(field)
    a.append(": ")
    a.append(data[i])
    a.append("\n")
  a.append(sep)
  a.append(data[BODY_L])
  return "".join(a)


# lazily loaded mailfiles
class MailMgr:
  def __init__(self):
    self.ids = msgids.Ids()
    # filename to data array
    # None or not present if data is not loaded
    self.datas = {}
    
  def ensure_loaded(self, fname):
    if fname not in self.datas or self.datas[fname] is None:
      self.datas[fname] = read_file(fname)

        
  def get(self, fname, ind):
    self.ensure_loaded(fname)
    return self.datas[fname][ind]

  def get_all(self, fname):
    self.ensure_loaded(fname)
    return self.datas[fname]

  def set(self, fname, ind, val):
    self.ensure_loaded(fname)
    self.datas[fname][ind] = val
    write_to_file(fname, self.datas[fname])

  def get_attachlist(self, fname):
    self.ensure_loaded(fname)
    s = self.datas[fname][ATTACH_L]
    return attach.attach_str_to_paths(s)

  def set_attachlist(self, fname, pathlist):
    self.set(fname, ATTACH_L, attach.attach_paths_to_str(pathlist))

  def get_references(self, fname):
    my_id = self.get(fname, MSG_ID_L).strip()
    refstr = self.get(fname, REFS_L)
    return do_get_referencelist(my_id, refstr)

  # get msgids that I am referenced by
  def get_refby(self, fname):
    s = self.get(fname, REF_BY_L)
    return do_get_referencelist(None, s)

  def add_refby(self, fname, refd_by):
    prev_refbys = set(self.get_refby(fname))
    prev_refbys.add(refd_by)
    self.set(fname, REF_BY_L, ", ".join(sorted(list(prev_refbys))))

  # get the id of the message that 'fname' replies to, or None
  def get_parent_id(self, fname):
    return do_get_parent_id(self.get(fname, MSG_ID_L), self.get_references(fname))

  def get_tags(self, fname, include_folders=False):
    return tags.get_taglist_from_str(self.get(fname, TAGS_L), include_folders=include_folders)

  def get_tagset(self, fname, include_folders=False):
    return tags.get_tagset_from_str(self.get(fname, TAGS_L), include_folders=include_folders)

  def add_tags(self, flist, my_taglist):
    taglist = tags.get_taglist_from_list(my_taglist)  # clean, sort tags
    if len(taglist) > 0:
      for fname in flist:
        new_taglist = tags.join(taglist, self.get_tags(fname))
        self.set(fname, TAGS_L, tags.get_str_from_clean_taglist(new_taglist))

  def remove_tags(self, flist, my_taglist):
    taglist = tags.get_taglist_from_list(my_taglist)  # clean, sort tags
    if len(taglist) > 0:
      for fname in flist:
        new_taglist = tags.get_nonmatching(self.get_tags(fname), taglist)
        self.set(fname, TAGS_L, tags.get_str_from_clean_taglist(new_taglist))

  def set_tags(self, flist, taglist):
    tagstr = tags.get_str_from_taglist(taglist)  # clean tags
    for fname in flist:
      self.set(fname, TAGS_L, tagstr)

  # if there was an update to these files (added, changed, or deleted),
  # just remove them from dict and re-load at next request
  def updated_flist(self, flist):
    for f in flist:
      if f in self.datas:
        del self.datas[f]

  def create(self, flist, datalist):
    idlist = [d[MSG_ID_L] for d in datalist]
    self.ids.add(idlist, flist)
    for fname,d in zip(flist, datalist):
      self.datas[fname] = d
      write_to_file(fname, d)
    # now that they're all written, update referenced-by info
    for fname in flist:
      for i_ref in self.get_references(fname):
        if self.ids.has(i_ref):
          self.add_refby(self.ids.get(i_ref), self.get(fname, MSG_ID_L))

  # when files are deleted
  def update_for_trashed(self, flist, idlist):
    for fname in flist:
      if fname in self.datas:  # which it should be...
        del self.datas[fname]
    self.ids.remove(idlist, flist)


def main_cli():
  if len(sys.argv) <= 1 or "-h" in sys.argv:
    print(usage_str)
    exit(0)
  flist = [line.strip() for line in sys.stdin.readlines()]
  mgr = MailMgr()
  if sys.argv[1] == "-g":
    if len(sys.argv) <= 2:
      print(usage_str)
      exit(0)
    if sys.argv[2] == "all":
      print("\n\n".join([data_to_str(mgr.get_all(fname)) for fname in flist]))
    else:
      inds = [fields.index(s) for s in sys.argv[2:]]
      print("\n\n".join(["\n".join([mgr.get(fname, i) for i in inds]) for fname in flist]))
  elif sys.argv[1] == "-add-tags":
    taglist = sys.argv[2:]
    mgr.add_tags(flist, taglist)
  elif sys.argv[1] == "-rm-tags":
    taglist = sys.argv[2:]
    mgr.remove_tags(flist, taglist)
  else:
    print(usage_str)


if __name__ == "__main__":
  main_cli()




####
# bomail.config.sendconfig
#
# Main configuration file.
# When loaded, it tries to open config file and read it; if unsuccessful,
# it runs a short setup program asking questions.
####

import os, sys

from bomail.config.conf_setup import options_dict


# --------------------------------------------------------------
# User info

names = []
email_addrs = []
smtp_servernames = []
smtp_ports = []
smtp_usernames = []
smtp_passwords = []

# return True on success, False on failure
# for user 1, there's no suffix e.g. name = MY_NAME
# for users 2 on, there are, e.g. name2 = MY_OTHER_NAME
def get_user_info(j):
  suffix = "" if j==0 else str(j)
  if not ("name" + suffix) in options_dict:
    return False
  name = options_dict["name" + suffix]
  email_addr = options_dict["email_addr" + suffix]
  servname = options_dict["smtp_servername" + suffix]
  if servname == "localhost":
    # no need to get password, port, etc
    smtp_servernames.append(servname)
    names.append(name)
    email_addrs.append(email_addr)
    return True
  smtp_ports.append(int(options_dict["smtp_port" + suffix]))
  if ("smtp_userpass_file" + suffix) in options_dict:
    try:
      with open(options_dict["smtp_userpass_file" + suffix]) as f:
        for line in f:
          if line.startswith("username ="):
            uname = line[line.index("=")+1:].strip()
          if line.startswith("password ="):
            pword = line[line.index("=")+1:].strip()
    except:
      with open(error_log_file, "a") as f:
        f.write("config: Could not read username/password from " + options_dict["smtp_userpass_file" + suffix])
      return False
  else: # get username and password straight from the config
    uname = options_dict["smtp_username" + suffix]
    pword = options_dict["smtp_password" + suffix]

  smtp_servernames.append(servname)
  names.append(name)
  email_addrs.append(email_addr)
  smtp_usernames.append(uname)
  smtp_passwords.append(pword)
  return True

# first get myname, then myname2, myname3, ...
get_user_info(0)
j = 2
while get_user_info(j):
  j += 1


my_aliases = []
if "alias_addresses" in options_dict:
  my_aliases = [s.strip() for s in options_dict["alias_addresses"].split(",")]






import os, sys

# this file checks for the configuration file
# if not found, it runs a setup and exists
# otherwise, it parses it into options_dict

# save location of config file (and old config file)
old_config_filename = ".bomailrc"
config_filename = "bomail.conf"

home = os.getenv("HOME")

# pre-0.9.3 versions
old_config_file = os.path.join(home, old_config_filename)

# post-0.9.3 versions
general_config_dir = os.getenv("XDG_CONFIG_HOME")
if general_config_dir is None or len(general_config_dir) == 0:
  general_config_dir = os.path.join(home, ".config")  # XDG spec
config_dir = general_config_dir

config_file = os.path.join(config_dir, config_filename)


# Check for config file, if found return
# otherwise, run setup and exit(0)
if not os.path.exists(config_file) and not os.path.exists(old_config_file):
  # do setup!
  general_data_dir = os.getenv("XDG_DATA_HOME")
  if general_data_dir is None or len(general_data_dir) == 0:
    general_data_dir = os.path.join(home, ".local", "share")  # XDG spec
  default_datadir = os.path.join(general_data_dir, "bomail")
  
  default_newmail_dir = os.path.join(home, "mail/new")
  default_oldmail_dir = os.path.join(home, "mail/cur")
  
  print("\nWelcome to bomail!")
  sys.stdout.write("Name (as it will appear in From line): ")
  sys.stdout.flush()
  myname = sys.stdin.readline().strip()
  sys.stdout.write("Email address (primary, i.e. send email from): ")
  sys.stdout.flush()
  myaddr = sys.stdin.readline().strip()

  sys.stdout.write("Email data location? (leave blank to use " + default_datadir +"): ")
  sys.stdout.flush()
  datadir = sys.stdin.readline().strip()
  if len(datadir) == 0:
    datadir = default_datadir

  sys.stdout.write("Process new mail from where? (leave blank to use " + default_newmail_dir + "): ")
  sys.stdout.flush()
  newmail_dir = sys.stdin.readline().strip()
  if len(newmail_dir) == 0:
    newmail_dir = default_newmail_dir

  sys.stdout.write("Put mail where after processing? (leave blank to use " + default_oldmail_dir + "): ")
  sys.stdout.flush()
  oldmail_dir = sys.stdin.readline().strip()
  if len(oldmail_dir) == 0:
    oldmail_dir = default_oldmail_dir

  sys.stdout.write("Add bomail command tab-completion to your ~/.bashrc file? (y/n): ")
  sys.stdout.flush()
  tabcomplete = sys.stdin.readline().strip()
  if len(tabcomplete) > 0 and tabcomplete[0] in ["y", "Y"]:
    with open(os.path.join(home, ".bashrc"), "a") as f:
      f.write(conf_strs.bomail_tabcomplete_str)

  # this is not ideal because it doesn't change if the configuration at
  # the bottom of the file changes
  metadata_dir  = os.path.join(datadir, "metadata/")
  handlers_file = os.path.join(metadata_dir, "mail-handlers.txt")
  if not os.path.exists(handlers_file):
    os.makedirs(metadata_dir, exist_ok=True)
    with open(handlers_file, "w") as f:
      f.write(conf_strs.sample_mailhandlers_str)

  s = conf_strs.sample_config_str.replace("YOUR NAME", myname).replace("USER@DOMAIN.COM", myaddr).replace("DATADIR", datadir).replace("NEWMAILDIR",newmail_dir).replace("OLDMAILDIR",oldmail_dir).replace("HOMEDIR",home)
  os.makedirs(config_dir, exist_ok=True)
  with open(config_file, "w") as f:
    f.write(s)

  print("""
Basic setup successful!
 (1) Edit """ + config_file + """ in order to send email and other options.
 (2) Edit """ + handlers_file + """ for filtering new mail.

Run bomail to begin...
""")
  exit(0)

# migrate from pre-0.9.3 old versions of bomail to 0.9.3
if not os.path.exists(config_file): # then old_config_file exists
  os.makedirs(config_dir, exist_ok=True)
  with open(old_config_file) as f:
    s = f.read()
  with open(config_file, "w") as f:
    f.write(s)
  os.remove(old_config_file)

# Function to open and parse config file
def parse_config_file():
  options_dict = {}
  def parse_line(line):
    line = line.strip()
    if len(line) > 0 and line[0] != "#":
      if "=" in line:
        left, right = line.split("=")
        options_dict[left.strip()] = right.strip()
      else:
        options_dict[line.strip()] = True
  
  with open(config_file) as f:
    for line in f:
      parse_line(line)
  return options_dict


# Actually do it
options_dict = parse_config_file()

options_dict["config_file"] = config_file




####
# bomail.guistuff.tabthread
#
# Tabs when configuration has threading, i.e.
# emails are grouped into conversations.
####

import shlex
import curses

from bomail.config.config import guicfg

import bomail.cli.mailfile as mailfile
import bomail.cli.search as search

import bomail.util.datestuff as datestuff
import bomail.util.merge_lines as merge_lines
import bomail.util.util as util

import bomail.guistuff.display as display
import bomail.guistuff.display_fmt as display_fmt
import bomail.guistuff.threadreader as threadreader
from bomail.guistuff.viewinterface import ViewInterface

# bound num between 0 and len(mylist)-1,
# unless the list is empty in which case 0
def bound(num, mylist):
  return max(0, min(num, len(mylist)-1))

class TabThread(ViewInterface):
  def __init__(self, search_str, gui):
    self.search_str = search_str
    self.gui = gui
    self.is_loaded = False
    self.cursor_ind = 0
    self.cursor_key = None  # where it sorted
    self.display_ind = 0
    self.num_marked = 0
    self.num_displayed = 0

    # list of filenames matching our search query
    # and total (pulled in by threads)
    self.matching_files = []

    # list of [match_list, all_list, is_marked, display_data]
    # where match_list is all the files in the thread matching query
    # and all_list is all files in the thread, both sorted
    self.thread_info = []

    # for sorting lists and thread_info
    self.sort_new = not "-sortold" in shlex.split(self.search_str)
    self.file_key = lambda a: a  #datestuff.parse_to_utc_datestr(self.gui.mail_mgr.get(a, mailfile.DATE_L))
    self.filelist_key = lambda l: self.file_key(l[0])
    self.tup_key = lambda t: self.filelist_key(t[0])


  def load(self):
    display.draw_loading_screen(self.gui)
    self.matching_files = search.search_argstr(self.search_str, self.gui.mail_mgr)
    self.matching_files.sort(key=self.file_key, reverse=self.sort_new)
    self.cursor_ind = 0
    self.cursor_key = None  # where it sorted
    self.display_ind = 0
    self.num_marked = 0
    self.num_displayed = 0
    # list of (matching, root_container, all) for each thread
    self.thread_info = []
    self.reload_thread_info()
    self.is_loaded = True

  def mark_obsolete(self):
    self.is_loaded = False

  def check_bounds(self):
    self.cursor_ind = bound(self.cursor_ind, self.thread_info)
    self.display_ind = min(self.display_ind, self.cursor_ind)
    self.get_draw_info(self.gui)
    if self.display_ind + self.num_displayed < self.cursor_ind:
      self.display_ind = self.cursor_ind - self.num_displayed - 1
    self.display_ind = bound(self.display_ind, self.thread_info)
    self.cursor_key = None if len(self.thread_info) == 0 else self.tup_key(self.thread_info[self.cursor_ind])

  # get files in currently marked threads, or under cursor
  def get_curr_filenames(self):
    if not self.is_loaded:
      self.load()
    if len(self.thread_info) == 0:
      return []
    elif self.num_marked == 0:
      return self.thread_info[self.cursor_ind][0]
    else:  # all files in all marked threads
      res = []
      for tup in self.thread_info:
        if tup[2]:  # is marked
          res.extend(tup[1])  # all files
      return res

  # using self.matching_files, get threads
  # and set cursor_index attempting to match previous location
  def reload_thread_info(self):
    oldsize = len(self.thread_info)
    # get new thread triples
    thread_trips = self.gui.thread_mgr.get_threads_for(self.matching_files, self.gui.mail_mgr)
    self.thread_info = [[trip[0], trip[2], False, None] for trip in thread_trips]
    self.cursor_ind = 0
    if self.cursor_key is not None:
      putbefore = (lambda t: t > self.cursor_key) if self.sort_new else (lambda t: self.cursor_key > t)
      for tup in self.thread_info:
        if putbefore(self.tup_key(tup)):
          self.cursor_ind += 1
        else:
          break
    self.num_marked = 0
    self.check_bounds()


  # new files have been imported into the system
  def update_for_new(self, filelist):
    if not self.is_loaded:
      return
    # the challenge is some of these files may not match, but be
    # replies to existing matches. It also changes sort order.
    # So re-request all threads
    new_matching = search.filter_argstr(self.search_str, self.gui.mail_mgr, filelist)
    new_matching.sort(key=self.file_key, reverse=self.sort_new)
    self.matching_files, temp = merge_lines.merge_lists(self.matching_files, new_matching, self.file_key, 0, reverse=self.sort_new)
    self.reload_thread_info()

  def update_for_change(self, filelist):
    if not self.is_loaded:
      return
    # build threads for all files that (a) did match and didn't change, (b) changed and now match
    fileset = set(filelist)
    changed_matching = search.filter_argstr(self.search_str, self.gui.mail_mgr, filelist)
    changed_matching.sort(key=self.file_key, reverse=self.sort_new)
    unchanged_matching = [f for f in self.matching_files if f not in fileset]
    oldlen = len(self.matching_files)
    self.matching_files, temp = merge_lines.merge_lists(unchanged_matching, changed_matching, self.file_key, 0, reverse=self.sort_new)
    self.reload_thread_info()

  # return True if we have no files left
  def update_for_trash(self, filelist):
    if not self.is_loaded:
      return True
    fileset = set(filelist)
    self.matching_files = [f for f in self.matching_files if f not in fileset]
    self.reload_thread_info()
    return len(self.thread_info) == 0

  # return lines, attr, is_unread, is_marked
  def get_disp_info(self, ind, width):
    info = self.thread_info[ind]
    if info[3] is None:  # display info not present
      info[3] = display_fmt.get_msg_lines_thread(self.gui.mail_mgr, info[0], info[1], width)
    is_unread = any([f in self.gui.unread_set for f in info[1]])
    is_marked = info[2]
    return info[3][0], info[3][1], is_unread, is_marked

  # return curr_view, mode, note
  # mode is "all", "view", or "note"
  # curr_view is self or a thread view to switch to
  def process_key(self, key):
    if not self.is_loaded:
      self.load()
    mode, note = "all", ""
    
    ## Navigation
    if key == guicfg.DOWN_KEY:
      self.scroll_down()
      mode, note = "view", ""
    if key == guicfg.UP_KEY:
      self.scroll_up()
      mode, note = "view", ""
    if key == guicfg.WAY_DOWN_KEY:
      self.scroll_down(amt=10)
      mode, note = "view", ""
    if key == guicfg.WAY_UP_KEY:
      self.scroll_up(amt=10)
      mode, note = "view", ""

    ## Mark/unmark threads
    elif key == guicfg.MARK_KEY:
      curr_info = self.thread_info[self.cursor_ind]
      name = display_fmt.get_shortened(curr_info[0][0], self.gui.mail_mgr)
      if curr_info[2]:  # is marked
        curr_info[2] = False
        self.num_marked -= 1
        mode, note = "view", "Un-marked " + name
      else:
        curr_info[2] = True
        self.num_marked += 1
        mode, note = "view", "Marked " + name
    elif key == guicfg.MARK_ALL_KEY:
      if self.num_marked == len(self.thread_info):  # unmark all
        for info in self.thread_info:
          info[2] = False
        self.num_marked = 0
        mode, note = "view", "Marked none"
      else:
        for info in self.thread_info:
          info[2] = True
        self.num_marked = len(self.thread_info)
        mode, note = "view", "Marked all"
     
    ## Viewing
    elif key == guicfg.VIEW_KEY or key == guicfg.RIGHT_KEY:
      myinfo = self.thread_info[self.cursor_ind]
      self.gui.mark_read(myinfo[1])
      return threadreader.ThreadReader(self.gui, myinfo[1]), "all", ""

    return self, mode, note


  def scroll_up(self, amt=1):
    buff = max(0, int((self.num_displayed-2)/2))
    if self.cursor_ind - self.display_ind <= buff:
      self.display_ind -= amt
    self.cursor_ind -= amt
    self.check_bounds()

  def scroll_down(self, amt=1):
    buff = max(0, int((self.num_displayed-2)/2))
    if self.display_ind + self.num_displayed-1 - self.cursor_ind <= buff:
      self.display_ind += amt
    self.cursor_ind += amt
    self.check_bounds()

  # return lines_of_txt, attribute_data
  # this is verbatim tabnothread except for file_info/thread_info...
  # probably can combine them...
  def get_draw_info(self, gui):
    height, width = gui.tab_area.getmaxyx()
    ind = self.display_ind
    curr_y = 0
    self.num_displayed = 0
    all_lines = []
    all_attr_data = []
    while curr_y < height and ind < len(self.thread_info):
      self.num_displayed += 1
      msg_lines, orig_attr_data, is_unread, is_marked = self.get_disp_info(ind, width)
      attr_data = list(orig_attr_data)
      if is_unread:
        attr_data = [(y, x, w, a | curses.A_BOLD if y==0 else a) for y,x,w,a in attr_data]
      if is_marked:
        all_lines.append(msg_lines[0][:-2] + "X" + msg_lines[0][-1:])
        all_lines.extend(msg_lines[1:])
      else:
        all_lines.extend(msg_lines)
      if ind == self.cursor_ind:
        # highlight the first line
        attr_data = [(y, x, w, a | guicfg.cursor_attr if y==0 else a) for y,x,w,a in attr_data]
      
      # add this message info to the list
      all_attr_data.extend([(curr_y+y, x, w, a) for y,x,w,a in attr_data])
      curr_y += len(msg_lines)
      ind += 1
    return all_lines, all_attr_data

  def mod_note(self, note, width):
    num_threads = len(self.thread_info)
    count_note = "No threads" if num_threads == 0 else "Thread " + str(self.cursor_ind+1) + " of " + str(num_threads)
    gap = width - (len(note) + len(count_note) + 4)
    if gap >= 0:
      note += " "*(gap + 3) + count_note
    return note
  

  # also verbatim from tabnothread
  def redraw(self, gui):
    if not self.is_loaded:
      self.load()
    # save cursor key
    height, width = gui.tab_area.getmaxyx()
    all_lines, all_attr_data = self.get_draw_info(gui)
    display.write_lines(gui.tab_area, 0, height, 0, 0, all_lines, all_attr_data)
    gui.tab_area.refresh()





####
# bomail.guistuff.gui_util
#
# various UI actions, like Send, that get called from several places.
####

import os
import subprocess
import curses

from bomail.config.config import pathcfg,guicfg

import bomail.cli.compose as compose
import bomail.cli.send as send
import bomail.cli.mailfile as mailfile

import bomail.guistuff.display as display
import bomail.guistuff.display_fmt as display_fmt
import bomail.guistuff.get_txt as get_txt

import bomail.util.addr as addr
import bomail.util.attach as attach
import bomail.util.datestuff as datestuff
import bomail.util.util as util


# return mode, note
def go_send(gui, filelist):
  if len(pathcfg.smtp_servernames) == 0:
    return "note", "Must configure ~/.bomailrc to send email"
  for filename in filelist:
    attach_list = gui.mail_mgr.get_attachlist(filename)
    if len(attach_list) > 0:
      for f in attach_list:
        if not (os.path.exists(f) and os.path.isfile(f)):
          return "all", "Error (did not send): could not find an attachment: " + str(f)
  lines = ["y  yes, send the email",
           "n  no, cancel"]
  if len(filelist) > 1:
    lines[0] = "y  yes, send " + str(len(filelist)) + " emails"
  display.draw_bottom_text(gui, lines)
  display.redraw_note(gui, "")
  key = gui.screen.getkey()
  if key == "y":
    display.redraw_note(gui, "Sending...")
    results, err_str = send.main(filelist, gui.mail_mgr)
    # trash all that were sent
    done = [fname for res,fname in zip(results,filelist) if res]
    if len(done) > 0:
      gui.acts.exec_act(("trash", done, [gui.mail_mgr.get_all(filename) for filename in done]))
    if len(done) == len(filelist):
      return "all", "Sent mail!"
    else:
      util.err_log("Sending mail: could not send: " + ", ".join([fname for res,fname in zip(results,filelist) if not res]) + "\n")
      util.err_log(err_str)
      return "all", "Some errors, sent " + str(len(done)) + " of " + str(len(filelist)) + " : " + err_str
  else:
    return "all", "Cancelled, did not send"


# return mode, note
def go_compose_draft(gui, key, filename, quote=True):
  if key == "n" or filename is None:
    recip_pairs = get_txt.get_recips(gui.tab_area, gui.addr_book)

    gui.tab_area.clear()
    gui.tab_area.refresh()
    subj = get_txt.get_subj_line(gui, "")
    if subj is None:
      return "all", "Cancelled"
    new_filename = compose.new_compose(subj, addr.pairlist_to_str(recip_pairs), "", "", "", "", gui.mail_mgr)
  elif key == "a":
    new_filename = compose.reply_all(filename, gui.mail_mgr, quote)
  elif key == "o":
    new_filename = compose.reply_one(filename, gui.mail_mgr, quote)
  elif key == "r":
    new_filename = compose.reply_all(filename, gui.mail_mgr, quote, include_sender=False)
  elif key == "f":
    recip_pairs = get_txt.get_recips(gui.tab_area, gui.addr_book)
    new_filename = compose.forward(filename, addr.pairlist_to_str(recip_pairs), gui.mail_mgr, quote)
  elif key == "e":
    new_filename = filename
  else:
    return "all", "Cancelled"
  subprocess.call(guicfg.edit_prog + " \"" + new_filename + "\"", shell=True)
  gui.reset_after_prog()
  gui.mail_mgr.updated_flist([new_filename])
  gui.update_for_change([new_filename])
  return "all", "Edited " + display_fmt.get_shortened(new_filename, gui.mail_mgr)


# if the WRITE key was pressed, get the kind of draft to create
# given current filename
# return mode, note, new_filename
def go_write_key(gui, filename):
  if filename[-5:] == "draft":
    return go_write_key_draft(gui, filename)
  else:
    return go_write_key_email(gui, filename)


def go_write_key_email(gui, filename):
  lines = [
    "a  reply-all        f  forward",
    "o  reply-one        n  new draft",
    "r  reply-recipient  z  CANCEL"]
  display.draw_bottom_text(gui, lines)
  display.redraw_note(gui, "")
  key = gui.screen.getkey()
  if key in ["a", "o", "r", "f", "n"]:
    return go_compose_draft(gui, key, filename)
  else:
    return "all", "Cancelled"


def go_write_key_draft(gui, filename):
  lines = [
    "n  new draft          t/y  add/remove recip  a/x  add/remove attachment",
    "e  edit this draft    c/v  add/remove CC       z  CANCEL",
    "s  edit subject line  b/m  add/remove BCC"]
  display.draw_bottom_text(gui, lines)
  display.redraw_note(gui, "")
  key = gui.screen.getkey()
  mode, text = "all", "Cancelled"
  if key in ["n", "e"]:
    mode, text = go_compose_draft(gui, key, filename)
  
  elif key == "s":
    s = get_txt.get_subj_line(gui, gui.mail_mgr.get(filename, mailfile.SUBJ_L))
    if s is None:
      return "all", "Cancelled"
    gui.mail_mgr.set(filename, mailfile.SUBJ_L, s)
    mode, text = "all", "Edited subject line of " + display_fmt.get_shortened(filename, gui.mail_mgr)

  elif key == "a":
    s = get_txt.get_path(gui.tab_area, gui.mail_mgr)
    if s is None:
      return "all", "Cancelled"
    old_attachlist = gui.mail_mgr.get_attachlist(filename)
    mode, text = gui.acts.do(("add attach", [filename], old_attachlist, [f.strip() for f in s]))

  elif key == "x":
    old_attachlist = gui.mail_mgr.get_attachlist(filename)
    if len(old_attachlist) == 0:
      return "all", "No attachments to remove"
    s = get_txt.get_sublist_from_list(gui.tab_area, old_attachlist, [], allow_new=False)
    if s is None:
      return "all", "Cancelled"
    mode, text = gui.acts.do(("remove attach", [filename], old_attachlist, s))

  else:  # recipient stuff
    if key in ["t", "y"]:
      mail_line = mailfile.TO_L
    elif key in ["c", "v"]:
      mail_line = mailfile.CC_L
    elif key in ["b", "m"]:
      mail_line = mailfile.BCC_L
    else:
      return "all", "Cancelled"
  
    # get the recipients and, if nonempty, do it
    old_recip_line = gui.mail_mgr.get(filename, mail_line)
    old_recip_pairs = addr.str_to_pairlist(old_recip_line)
    if key in ["t", "c", "b"]:
      # get new recips, add to mail_line
      new_recip_pairs = get_txt.get_recips(gui.tab_area, gui.addr_book, already_recip_pairs=old_recip_pairs)
      if len(new_recip_pairs) == 0:
        return "all", "Cancelled"
      new_recip_line = addr.pairlist_to_str(old_recip_pairs + new_recip_pairs)
    else:
      # get from old recips, remove from mail_line
      remove_recip_pairs = get_txt.get_recips(gui.tab_area, gui.addr_book, recip_pairs=old_recip_pairs, allow_new=False)
      if len(remove_recip_pairs) == 0:
        return "all", "Cancelled"
      new_recip_line = addr.pairlist_to_str([r for r in old_recip_pairs if r not in remove_recip_pairs])
    mode, text = gui.acts.do(("edit file", filename, mail_line, new_recip_line, old_recip_line))
  gui.mail_mgr.updated_flist([filename])
  gui.update_for_change([filename])
  return mode, text


# return mode, note
def go_schedule(gui, filelist):
  sched_txt = get_txt.get_sched_txt(gui)
  if sched_txt is None:
    return "view", "Cancelled"
  schedobj = datestuff.parse_schedstr_to_utcobj(sched_txt)
  if schedobj is None:
    return "view", "Cancelled: could not parse input"
  return gui.acts.do(("scheduled", filelist, [gui.mail_mgr.get(f, mailfile.STATE_L) for f in filelist], schedobj))


def go_add_tags(gui, filelist):
  prevtags = [gui.mail_mgr.get_tags(f) for f in filelist]
  oldtaglist = sorted(list(set.intersection(*[gui.mail_mgr.get_tagset(f) for f in filelist])))
  newtags = get_txt.get_tags(gui.tab_area, [t for t in gui.tag_mgr.tags if t not in oldtaglist])
  if newtags is None or len(newtags) == 0:
    return "view", "Cancelled"
  else:
    alltags = set(gui.tag_mgr.tags)
    unknown = [t for t in newtags if t not in alltags]
    return gui.acts.do(("add tags", filelist, newtags, prevtags, unknown))


def go_remove_tags(gui, filelist):
  prevtags = [gui.mail_mgr.get_tags(f) for f in filelist]
  oldtaglist = sorted(list(set.union(*[gui.mail_mgr.get_tagset(f) for f in filelist])))
  removeme = get_txt.get_tags(gui.tab_area, oldtaglist, allow_new=False)
  if removeme is None or len(removeme) == 0:
    return "view", "Cancelled"
  else:
    return gui.acts.do(("remove tags", filelist, removeme, prevtags))






# bomail.config.optconfig
# Options not in bomailrc (may edit)


# how many new emails to process at a time, -1 for no limit
process_new_limit = -1

# how many new emails to process in a batch, decrease to save memory
# for very large imports
process_batch_size = 1000

# tag drafts with same tags as parent
autotag_draft_replies = True

# Replying to emails
# will replace %from, %to, %date, %time, %body
# with the corresponding text from the quoted email
quote_format = """\n
%from wrote on %date at %time:
%body"""

# %a is day of week; %Y,%m,%d are year/month/date
# %H,%M are hour/minute
quote_date_fmt = "%a, %Y-%m-%d"
quote_time_fmt = "%H:%M"
quote_line_prefix = "> "




####
# bomail.config.config
#
# Main configuration file.
# When loaded, it tries to open config file and read it; if unsuccessful,
# it runs a short setup program asking questions.
####

# use: from bomail.config.config import guicfg,sendcfg,optcfg,pathcfg

import os, sys

# check for config file; if not, run setup and exit
from bomail.config.conf_setup import options_dict
import bomail.config.guiconfig as guicfg
import bomail.config.sendconfig as sendcfg
import bomail.config.optconfig as optcfg
import bomail.config.pathconfig

class PathConfig:
  def __init__(self):
    pass
pathcfg = PathConfig()

# loads pathcfg
def load_paths(new_data_base):
  bomail.config.pathconfig.get_paths(pathcfg,
    options_dict["config_file"],
    new_data_base,
    options_dict["new_rawmail_location"],
    options_dict["processed_rawmail_location"])

load_paths(options_dict["data_location"])





####
# bomail.guistuff.tabnothread
#
# Tabs when configuration turns off threading, i.e.
# each email gets its own line.
####

import os
import sys
import subprocess
import curses
import shlex
import functools

from bomail.config.config import guicfg

import bomail.cli.mailfile as mailfile
import bomail.cli.search as search
import bomail.cli.chstate as chstate
import bomail.cli.send as send
import bomail.cli.compose as compose

import bomail.util.addr as addr
import bomail.util.datestuff as datestuff
import bomail.util.util as util
import bomail.util.merge_lines as merge_lines

import bomail.guistuff.display as display
import bomail.guistuff.display_fmt as display_fmt
import bomail.guistuff.gui_util as gui_util
from bomail.guistuff.viewinterface import ViewInterface

# bound num between 0 and len(mylist)-1,
# unless the list is empty in which case 0
def bound(num, mylist):
  return max(0, min(num, len(mylist)-1))


class TabNoThread(ViewInterface):
  def __init__(self, search_str, gui):
    self.search_str = search_str
    self.gui = gui
    self.is_loaded = False
    self.cursor_ind = 0
    self.display_ind = 0
    self.num_marked = 0
    self.num_displayed = 0

    # list of [filename, is_marked, display_data] for each filename
    # where display_data is either None or (lines_to_display, attr_data)
    self.file_info = []

    # for sorting filenames and file_info
    self.sort_new = not "-sortold" in shlex.split(self.search_str)
    self.file_key = lambda a: a #datestuff.parse_to_utc_datestr(self.gui.mail_mgr.get(a, mailfile.DATE_L))
    self.tup_key  = lambda t: self.file_key(t[0])


  # get all the filenames matching our search_str
  def load(self):
    display.draw_loading_screen(self.gui)
    filenames = search.search_argstr(self.search_str, self.gui.mail_mgr)
    self.cursor_ind = 0
    self.display_ind = 0
    self.num_marked = 0
    self.num_displayed = 0
    self.file_info = [[f, False, None] for f in filenames]
    self.file_info.sort(key=self.tup_key, reverse=self.sort_new)
    self.is_loaded = True

  def mark_obsolete(self):
    self.is_loaded = False

  def check_bounds(self):
    self.cursor_ind = bound(self.cursor_ind, self.file_info)
    self.display_ind = min(self.display_ind, self.cursor_ind)
    self.get_draw_info(self.gui)
    if self.display_ind + self.num_displayed < self.cursor_ind:
      self.display_ind = self.cursor_ind - self.num_displayed - 1
    self.display_ind = bound(self.display_ind, self.file_info)

  # return all filenames to be affected by the current action
  def get_curr_filenames(self):
    if not self.is_loaded:
      self.load()
    if len(self.file_info) == 0:
      return []
    elif self.num_marked == 0:
      return [self.file_info[self.cursor_ind][0]]
    else:
      return [t[0] for t in self.file_info if t[1]]  # marked


  def update_for_new(self, filelist):
    if not self.is_loaded:
      return
    new_filenames = search.filter_argstr(self.search_str, self.gui.mail_mgr, filelist)
    if len(new_filenames) == 0:
      return
    new_filenames.sort(key=self.file_key, reverse=self.sort_new)
    new_info = [[f, False, None] for f in new_filenames]
    self.file_info, self.cursor_ind = merge_lines.merge_lists(self.file_info, new_info, self.tup_key, self.cursor_ind, reverse=self.sort_new)
    self.check_bounds()
    
  # first, remove those no longer matching and reset display info
  # of those changed but matching
  # second, merge with newly-matching
  def update_for_change(self, filelist):
    if not self.is_loaded:
      return
    mine_now = search.filter_argstr(self.search_str, self.gui.mail_mgr, filelist)
    mine_now_set = set(mine_now)
    notmine_now_set = set([f for f in filelist if f not in mine_now_set])
    revised_info = []
    for i,tup in enumerate(self.file_info):
      if tup[0] in notmine_now_set:
        if tup[1]:
          self.num_marked -= 1
        if i < self.cursor_ind:
          self.cursor_ind -= 1
        continue
      if tup[0] in mine_now_set:
        tup[2] = None  # reset display info
        mine_now_set.remove(tup[0])
      revised_info.append(tup)
    new_info = [[f, False, None] for f in mine_now_set]
    self.file_info, self.cursor_ind = merge_lines.merge_lists(revised_info, new_info, self.tup_key, self.cursor_ind, reverse=self.sort_new)
    self.check_bounds()
  
  def update_for_trash(self, filelist):
    if not self.is_loaded:
      return True
    fset = set(filelist)
    revised_info = []
    for i,tup in enumerate(self.file_info):
      if tup[0] in fset:
        if tup[1]:
          self.num_marked -= 1
        if i < self.cursor_ind:
          self.cursor_ind -= 1
        continue
      revised_info.append(tup)
    self.file_info = revised_info
    self.check_bounds()
    return len(self.file_info) == 0


  # lazily load display data and return lines, attr, is_unread, is_marked
  # memoize: the data is None until requested, then is loaded
  def get_disp_info(self, ind, width):
    info = self.file_info[ind]
    if info[2] is None:  # display info not present
      info[2] = display_fmt.get_msg_lines_nothread(self.gui.mail_mgr, info[0], width)
    is_unread = info[0] in self.gui.unread_set
    is_marked = info[1]
    return info[2][0], info[2][1], is_unread, is_marked


  # return curr_view, mode, note
  # mode is "all", "view", or "note"
  # curr_view is always self because there is not thread view
  def process_key(self, key):
    if not self.is_loaded:
      self.load()
    mode, note = "all", ""
    
    ## Navgation
    if key == guicfg.DOWN_KEY:
      self.scroll_down()
      mode, note = "view", ""
    elif key == guicfg.UP_KEY:
      self.scroll_up()
      mode, note = "view", ""
    elif key == guicfg.WAY_DOWN_KEY:
      self.scroll_down(amt=10)
      mode, note = "view", ""
    elif key == guicfg.WAY_UP_KEY:
      self.scroll_up(amt=10)
      mode, note = "view", ""

    # mark/unmark messages
    elif key == guicfg.MARK_KEY:
      curr_data = self.file_info[self.cursor_ind]
      name = display_fmt.get_shortened(curr_data[0], self.gui.mail_mgr)
      if curr_data[1]:  # is marked
        curr_data[1] = False
        self.num_marked -= 1
        mode, note = "view", "Un-marked " + name
      else:  # is not marked
        curr_data[1] = True
        self.num_marked += 1
        mode, note = "view", "Marked " + name

    elif key == guicfg.MARK_ALL_KEY:
      if self.num_marked == len(self.file_info):  # unmark all
        for info in self.file_info:
          info[1] = False
        self.num_marked = 0
        mode, note = "view", "Marked none"
      else:
        for info in self.file_info:
          info[1] = True
        self.num_marked = len(self.file_info)
        mode, note = "view", "Marked all"

    ## Viewing
    elif key == guicfg.VIEW_KEY:
      self.gui.mark_read([self.file_info[self.cursor_ind][0]])
      subprocess.call(guicfg.read_prog + " '" + self.file_info[self.cursor_ind][0] + "'", shell=True)
      self.gui.reset_after_prog()
      mode, note = "all", ""

    return self, mode, note


  def scroll_up(self, amt=1):
    buff = max(0, int((self.num_displayed-2)/2))
    if self.cursor_ind - self.display_ind <= buff:
      self.display_ind -= amt
    self.cursor_ind -= amt
    self.check_bounds()

  def scroll_down(self, amt=1):
    buff = max(0, int((self.num_displayed-2)/2))
    if self.display_ind + self.num_displayed-1 - self.cursor_ind <= buff:
      self.display_ind += amt
    self.cursor_ind += amt
    self.check_bounds()

#  def scroll_down_page(self):
#    self.scroll_down(self.display_ind + self.num_displayed - self.cursor_ind)
#
#  # attempt to scroll up until the previous cursor message leaves the screen
#  def scroll_up_page(self):
#    old_cursor_ind = self.cursor_ind
#    while self.cursor_ind > 0:
#      self.cursor_ind -= 1
#      self.get_draw_info(self.gui)  # set num_displayed
#      # check if we've scrolled up a whole page yet
#      if self.cursor_ind + self.num_displayed <= old_cursor_ind:
#        break
#    self.display_ind = self.cursor_ind

  # return lines_of_txt, attribute_data
  def get_draw_info(self, gui):
    height, width = gui.tab_area.getmaxyx()
    ind = self.display_ind
    curr_y = 0
    self.num_displayed = 0
    all_lines = []
    all_attr_data = []
    while curr_y < height and ind < len(self.file_info):
      self.num_displayed += 1
      msg_lines, orig_attr_data, is_unread, is_marked = self.get_disp_info(ind, width)
      attr_data = list(orig_attr_data)
      if is_unread:
        attr_data = [(y, x, w, a | curses.A_BOLD if y==0 else a) for y,x,w,a in attr_data]
      if is_marked:
        all_lines.append(msg_lines[0][:-2] + "X" + msg_lines[0][-1:])
        all_lines.extend(msg_lines[1:])
      else:
        all_lines.extend(msg_lines)
      if ind == self.cursor_ind:
        # highlight the first line
        attr_data = [(y, x, w, a | guicfg.cursor_attr if y==0 else a) for y,x,w,a in attr_data]
      
      # add this message info to the list
      all_attr_data.extend([(curr_y+y, x, w, a) for y,x,w,a in attr_data])
      curr_y += len(msg_lines)
      ind += 1
    return all_lines, all_attr_data


  def mod_note(self, note, width):
    num_msgs = len(self.file_info)
    count_note = "No messages" if num_msgs == 0 else "Message " + str(self.cursor_ind+1) + " of " + str(num_msgs)
    gap = width - (len(note) + len(count_note) + 4)
    if gap >= 0:
      note += " "*(gap + 3) + count_note
    return note
  

  def redraw(self, gui):
    if not self.is_loaded:
      self.load()
    height, width = gui.tab_area.getmaxyx()
    all_lines, all_attr_data = self.get_draw_info(gui)
    display.write_lines(gui.tab_area, 0, height, 0, 0, all_lines, all_attr_data)
    gui.tab_area.refresh()


